<?php

declare(strict_types=1);

namespace Drupal\ai_provider_openrouter\Plugin\AiProvider;

use Drupal\ai\Attribute\AiProvider;
use Drupal\ai\Base\AiProviderClientBase;
use Drupal\ai\Enum\AiModelCapability;
use Drupal\ai\Enum\AiProviderCapability;
use Drupal\ai\Exception\AiQuotaException;
use Drupal\ai\Exception\AiRateLimitException;
use Drupal\ai\Exception\AiResponseErrorException;
use Drupal\ai\Exception\AiUnsafePromptException;
use Drupal\ai\OperationType\Chat\ChatInterface;
use Drupal\ai\OperationType\Embeddings\EmbeddingsInterface;
use Drupal\ai\OperationType\Chat\ChatInput;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\ai\OperationType\Chat\ChatOutput;
use Drupal\ai\OperationType\Chat\Tools\ToolsFunctionOutput;
use Drupal\ai\OperationType\Embeddings\EmbeddingsInput;
use Drupal\ai\OperationType\Embeddings\EmbeddingsOutput;
use Drupal\ai\Traits\OperationType\ChatTrait;
use Drupal\ai_provider_openrouter\Service\OpenRouterClient;
use Drupal\Component\Serialization\Json;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\DependencyInjection\Attribute\Autowire;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;

/**
 * Plugin implementation of the 'openrouter' AI provider.
 */
#[AiProvider(
  id: 'openrouter',
  label: new \Drupal\Core\StringTranslation\TranslatableMarkup('OpenRouter'),
)]
class OpenRouterProvider extends AiProviderClientBase implements ChatInterface, EmbeddingsInterface, ContainerFactoryPluginInterface {

  use ChatTrait;

  /**
   * @var \Drupal\ai_provider_openrouter\Service\OpenRouterClient
   */
  protected $client;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var static $instance */
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->client = $container->get('ai_provider_openrouter.client');
    $instance->logger = $container->get('logger.factory')->get('ai_provider_openrouter');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getCapabilities(): array {
    return [
      AiProviderCapability::CHAT,
      AiProviderCapability::EMBEDDINGS,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedCapabilities(): array {
    return [
      AiProviderCapability::StreamChatOutput,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function chat(ChatInput|array|string $input, string $model_id, array $tags = []): ChatOutput {
    try {
      // Normalize the input if needed.
      $chat_input = $input;
      if ($input instanceof ChatInput) {
        $chat_input = [];
        // Add a system role if wanted.
        if ($this->chatSystemRole) {
          // If it's o1 or o3 in it, we add it as a user message (similar to OpenAI provider).
          if (preg_match('/(o1|o3)/i', $model_id)) {
            $chat_input[] = [
              'role' => 'user',
              'content' => $this->chatSystemRole,
            ];
          }
          else {
            $chat_input[] = [
              'role' => 'system',
              'content' => $this->chatSystemRole,
            ];
          }
        }
        
        /** @var \Drupal\ai\OperationType\Chat\ChatMessage $message */
        foreach ($input->getMessages() as $message) {
          $content = [
            [
              'type' => 'text',
              'text' => $message->getText(),
            ],
          ];
          
          // Support for images in messages if the model supports it
          if (count($message->getImages())) {
            foreach ($message->getImages() as $image) {
              $content[] = [
                'type' => 'image_url',
                'image_url' => [
                  'url' => $image->getAsBase64EncodedString(),
                ],
              ];
            }
          }
          
          $new_message = [
            'role' => $message->getRole(),
            'content' => $content,
          ];

          // If it's a tools response.
          if ($message->getToolsId()) {
            $new_message['tool_call_id'] = $message->getToolsId();
          }

          // If we want the results from some older tools call.
          if ($message->getTools()) {
            $new_message['tool_calls'] = $message->getRenderedTools();
          }

          $chat_input[] = $new_message;
        }
      }
      elseif (is_array($input)) {
        $chat_input = $input;
      }
      else {
        $chat_input = [
          'model' => $model_id,
          'messages' => [['role' => 'user', 'content' => (string) $input]],
        ];
      }

      $payload = [
        'model' => $model_id,
        'messages' => $chat_input,
      ] + $this->configuration;
      
      // If we want to add tools to the input.
      if (is_object($input) && method_exists($input, 'getChatTools') && $input->getChatTools()) {
        $payload['tools'] = $input->getChatTools()->renderToolsArray();
        foreach ($payload['tools'] as $key => $tool) {
          $payload['tools'][$key]['function']['strict'] = FALSE;
        }
      }
      
      // Check for structured json schemas.
      if (is_object($input) && method_exists($input, 'getChatStructuredJsonSchema') && $input->getChatStructuredJsonSchema()) {
        $payload['response_format'] = [
          'type' => 'json_schema',
          'json_schema' => $input->getChatStructuredJsonSchema(),
        ];
      }


      $response = $this->client->chatCompletion($payload);
      
      // Process tools if present in the response
      $tools = [];
      if (!empty($response['choices'][0]['message']['tool_calls'])) {
        foreach ($response['choices'][0]['message']['tool_calls'] as $tool) {
          $arguments = Json::decode($tool['function']['arguments']);
          $tools[] = new ToolsFunctionOutput($input->getChatTools()->getFunctionByName($tool['function']['name']), $tool['id'], $arguments);
        }
      }
      
      $message = new ChatMessage(
        $response['choices'][0]['message']['role'],
        $response['choices'][0]['message']['content'] ?? "",
        []
      );
      
      if (!empty($tools)) {
        $message->setTools($tools);
      }
      
      return new ChatOutput($message, $response, []);
    }
    catch (\Exception $e) {
      // Try to figure out rate limit issues.
      if (strpos($e->getMessage(), 'Request too large') !== FALSE) {
        throw new AiRateLimitException($e->getMessage());
      }
      if (strpos($e->getMessage(), 'Too Many Requests') !== FALSE) {
        throw new AiRateLimitException($e->getMessage());
      }
      // Try to figure out quota issues.
      if (strpos($e->getMessage(), 'You exceeded your current quota') !== FALSE) {
        throw new AiQuotaException($e->getMessage());
      }
      // Handle unsafe content errors
      if (strpos($e->getMessage(), 'content policy violation') !== FALSE) {
        throw new AiUnsafePromptException($e->getMessage());
      }
      // Handle general API errors
      throw new AiResponseErrorException($e->getMessage(), $e->getCode(), $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function embeddings(EmbeddingsInput|string $input, string $model_id, array $tags = []): EmbeddingsOutput {
    try {
      if ($input instanceof EmbeddingsInput) {
        $model = $input->getModel() ?? $model_id;
        $embeddings_input = $input->getInput();
      }
      else {
        $model = $model_id;
        $embeddings_input = $input;
      }

      $result = $this->client->embeddings([
        'model' => $model,
        'input' => $embeddings_input,
        // Optionally pass $tags or other options if needed.
      ]);
      return new EmbeddingsOutput($result);
    }
    catch (\Exception $e) {
      // Try to figure out rate limit issues.
      if (strpos($e->getMessage(), 'Request too large') !== FALSE) {
        throw new AiRateLimitException($e->getMessage());
      }
      if (strpos($e->getMessage(), 'Too Many Requests') !== FALSE) {
        throw new AiRateLimitException($e->getMessage());
      }
      // Try to figure out quota issues.
      if (strpos($e->getMessage(), 'You exceeded your current quota') !== FALSE) {
        throw new AiQuotaException($e->getMessage());
      }
      // Handle general API errors
      throw new AiResponseErrorException($e->getMessage(), $e->getCode(), $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getConfig(): \Drupal\Core\Config\ImmutableConfig {
    // Return the module's config object.
    return \Drupal::config('ai_provider_openrouter.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getApiDefinition(): array {
    // Return an empty array or define the API schema if needed.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getModelSettings(string $model_id, array $generalConfig = []): array {
    // Return model-specific settings if needed, or an empty array.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function isUsable(?string $operation_type = NULL, array $capabilities = []): bool {
    $config = $this->getConfig();
    $api_key = $config->get('api_key');
    $usable = !empty($api_key);
    return $usable;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedOperationTypes(): array {
    // Support chat and embeddings.
    return ['chat', 'embeddings'];
  }

  /**
   * {@inheritdoc}
   */
  public function setAuthentication(mixed $authentication): void {
    // Stub: not used in this provider yet.
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailableConfiguration(string $operation_type, string $model_id): array {
    // Stub: no advanced config yet.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultConfigurationValues(string $operation_type, string $model_id): array {
    // Stub: no advanced config yet.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getInputExample(string $operation_type, string $model_id): mixed {
    // Stub: no example provided.
    return null;
  }

  /**
   * {@inheritdoc}
   */
  public function getAuthenticationExample(string $operation_type, string $model_id): mixed {
    // Stub: no example provided.
    return null;
  }

  /**
   * {@inheritdoc}
   */
  public function maxEmbeddingsInput($model_id = ''): int {
    // Return a reasonable default. Adjust as needed for OpenRouter models.
    return 8192;
  }

  /**
   * {@inheritdoc}
   * 
   * OpenRouter provides access to a wide range of models from different providers.
   * To improve usability, this implementation filters available models based on
   * user configuration in the OpenRouter settings form.
   * 
   * This approach allows site administrators to curate a focused list of models
   * they want to use, preventing dropdown menus throughout the site from being
   * overwhelmed with hundreds of options.
   * 
   * The filtering is based on the 'enabled_models' configuration, which stores
   * the IDs of models that the administrator has explicitly enabled in the
   * OpenRouter configuration form.
   *
   * @param string|null $operation_type
   *   The operation type to filter models by (e.g., 'chat', 'embeddings').
   * @param array $capabilities
   *   Optional array of capabilities to filter models by.
   *
   * @return array
   *   An array of model IDs and labels, filtered by user configuration.
   */
  public function getConfiguredModels(?string $operation_type = NULL, array $capabilities = []): array {
    $models = $this->client->listModels();
    $result = [];
    // Get enabled models from config, if any.
    $config = \Drupal::config('ai_provider_openrouter.settings');
    $enabled = $config->get('enabled_models') ?? [];
    $enabled = array_filter($enabled);

    $this->logger->debug('OpenRouter getConfiguredModels: Found @count total models, @enabled enabled', [
      '@count' => count($models),
      '@enabled' => count($enabled),
    ]);

    foreach ($models as $id => $model) {
      $label = $model['name'] ?? $model['id'];
      // If enabled models are set, only include those.
      if (empty($enabled) || in_array($id, $enabled, true)) {
        $result[$id] = $label;
      }
    }
    return $result;
  }
}



